{*** Sprite sorting routines. ***}

{But doesn't SAT sort automatically, unless you turn sorting off? Yes, it does, but there are}
{times when you know more than SAT concerning how to sort. SAT does its best, but it can't}
{know that you are moving around a whole pile of objects, or grab certain objects and need}
{them brought to front immediately; SAT can't waste the time to do a complete sort for}
{every frame.}
{}
{Use the SortOne routines if you have one sprite that needs to be perfectly sorted.}
{}
{If you want to sort the entire sprite list, use BucketSort. The actual sorting is extremely}
{fast, but does not keep track of what sprites were moved.}
{}
{Note: If you use SATRun2, you must set the "dirty" flags of any sprites that overlap and change}
{order, or you will get incorrect results! (SAT does this itself in the internal sorting system.)}
{}
{Note: Don't sort the sprite list from within a sprite handler! If you do, some sprite handler}
{calls may be called severeal times or not at all. Manipulation of the sprite list should}
{usually be done from the main loop.}

unit SortingUtils;

interface

	uses
{$ifc UNDEFINED THINK_PASCAL}
		Types, QuickDraw, Events, Windows, Dialogs, Fonts, DiskInit, TextEdit, Traps,{}
		Memory, SegLoad, Scrap, ToolUtils, OSUtils, Menus, Resources, StandardFile,{}
		GestaltEqu, Files, Errors, Devices, QuickDrawText, 
{$endc}
		SAT;

{Unlinks a sprite from the sprite list.}
	procedure UnlinkSprite (sp: SpritePtr);

{Puts a certain sprite at a sorted position according to layer.}
	procedure SortOne (sp: SpritePtr);
{Same thing using position.v:}
	procedure SortOneV (sp: SpritePtr);

{BucketSort is useful if you need to make sure the sprite list gets sorted at}
{once - and it is really fast!}
	procedure BucketSort (minLayer, maxLayer: Integer);
{Same thing using position.v:}
	procedure BucketSortV (minLayer, maxLayer: Integer);

implementation

	procedure UnlinkSprite (sp: SpritePtr);
	begin
		if sp = gSAT.sRoot then
			gSAT.sRoot := sp^.next;
		if sp^.next <> nil then
			sp^.next^.prev := sp^.prev;
		if sp^.prev <> nil then
			sp^.prev^.next := sp^.next;
		sp^.prev := nil;
		sp^.next := nil;
	end; {UnlinkSprite}

	procedure SortOne (sp: SpritePtr);
		var
			aSprite, next, prev, wasNext: SpritePtr;
	begin
		wasNext := sp^.next;
		UnlinkSprite(sp);

		aSprite := gSAT.sRoot;
		if aSprite <> nil then
			while (aSprite^.next <> nil) and (sp^.layer > aSprite^.layer) do
				aSprite := aSprite^.next;

		if aSprite = nil then
			begin
{Alone in the list! Put it in!}
				gSAT.sRoot := sp;
			end
		else if sp^.layer > aSprite^.layer then
			begin
{After aSprite!}
				sp^.next := aSprite^.next;	{torde vara nil!}
				sp^.prev := aSprite;
				aSprite^.next := sp;
				if sp^.next <> nil then
					sp^.next^.prev := sp;
			end
		else
			begin
{Before aSprite!}
				sp^.next := aSprite;
				sp^.prev := aSprite^.prev;
				aSprite^.prev := sp;
				if sp^.prev <> nil then
					sp^.prev^.next := sp
				else
					gSAT.sRoot := sp;
			end;
{If it has moved, set the dirty flag.}
		if sp^.next <> wasNext then
			sp^.dirty := true;
	end; {SortOne}

	procedure SortOneV (sp: SpritePtr);
		var
			aSprite, next, prev, wasNext: SpritePtr;
	begin
		wasNext := sp^.next;
		UnlinkSprite(sp);

		aSprite := gSAT.sRoot;
		if aSprite <> nil then
			while (aSprite^.next <> nil) and (sp^.position.v > aSprite^.position.v) do
				aSprite := aSprite^.next;

		if aSprite = nil then
			begin
{Alone in the list! Put it in!}
				gSAT.sRoot := sp;
			end
		else if sp^.position.v > aSprite^.position.v then
			begin
{After aSprite!}
				sp^.next := aSprite^.next;	{torde vara nil!}
				sp^.prev := aSprite;
				aSprite^.next := sp;
				if sp^.next <> nil then
					sp^.next^.prev := sp;
			end
		else
			begin
{Before aSprite!}
				sp^.next := aSprite;
				sp^.prev := aSprite^.prev;
				aSprite^.prev := sp;
				if sp^.prev <> nil then
					sp^.prev^.next := sp
				else
					gSAT.sRoot := sp;
			end;
{If it has moved, set the dirty flag.}
		if sp^.next <> wasNext then
			sp^.dirty := true;
	end; {SortOneV}

	procedure BucketSort (minLayer, maxLayer: Integer);
		type
			BucketArr = array[0..60000] of SpritePtr;
			BucketArrPtr = ^BucketArr;
		var
			sp, next: SpritePtr;
			buckets: BucketArrPtr;
			putInto: Integer;
			i: Integer;
	begin
		buckets := BucketArrPtr(NewPtrClear(SizeOf(Ptr) * (maxLayer - minLayer + 1)));
		if buckets = nil then
			Exit(BucketSort);

		sp := gSAT.sRoot;
		while sp <> nil do
			begin
				next := sp^.next;
{Unlink the sprite by removing old links.}
				sp^.prev := nil;
				sp^.next := nil;

				putInto := sp^.layer - minLayer;
				if putInto < 0 then
					putInto := 0;
				if putInto + minLayer > maxLayer then
					putInto := maxLayer - minLayer;

{Link into the proper bucket.}
				sp^.next := buckets^[putInto];
				if buckets^[putInto] <> nil then
					buckets^[putInto]^.prev := sp;
				buckets^[putInto] := sp;

				sp := next;
			end;

		gSAT.sRoot := nil;

{The sprite list is now empty, and all sprites are in the buckets!}

		next := nil;

		for i := 0 to maxLayer - minLayer do
			begin
{Link in the next bucket!}
				if next = nil then
					gSAT.sRoot := buckets^[i]
				else
					begin
						next^.next := buckets^[i];
						buckets^[i]^.prev := next;
					end;

{If the bucket wasn't empty, seach forwards to the last one.}
				if buckets^[i] <> nil then
					begin
						next := buckets^[i];
						while next^.next <> nil do
							next := next^.next;
					end;
			end;

		DisposePtr(Ptr(buckets));
	end; {BucketSort}

{Same thing but sorts on position.v instead!}
	procedure BucketSortV (minLayer, maxLayer: Integer);
		type
			BucketArr = array[0..60000] of SpritePtr;
			BucketArrPtr = ^BucketArr;
		var
			sp, next: SpritePtr;
			buckets: BucketArrPtr;
			putInto: Integer;
			i: Integer;
	begin
		buckets := BucketArrPtr(NewPtrClear(SizeOf(Ptr) * (maxLayer - minLayer + 1)));
		if buckets = nil then
			Exit(BucketSortV);

		sp := gSAT.sRoot;
		while sp <> nil do
			begin
				next := sp^.next;
{Unlink the sprite by removing old links.}
				sp^.prev := nil;
				sp^.next := nil;

				putInto := sp^.position.v - minLayer;			{<- only difference!}
				if putInto < 0 then
					putInto := 0;
				if putInto + minLayer > maxLayer then
					putInto := maxLayer - minLayer;

{Link into the proper bucket.}
				sp^.next := buckets^[putInto];
				if buckets^[putInto] <> nil then
					buckets^[putInto]^.prev := sp;
				buckets^[putInto] := sp;

				sp := next;
			end;

		gSAT.sRoot := nil;

{The sprite list is now empty, and all sprites are in the buckets!}

		next := nil;

		for i := 0 to maxLayer - minLayer do
			begin
{Link in the next bucket!}
				if next = nil then
					gSAT.sRoot := buckets^[i]
				else
					begin
						next^.next := buckets^[i];
						buckets^[i]^.prev := next;
					end;

{If the bucket wasn't empty, seach forwards to the last one.}
				if buckets^[i] <> nil then
					begin
						next := buckets^[i];
						while next^.next <> nil do
							next := next^.next;
					end;
			end;

		DisposePtr(Ptr(buckets));
	end; {BucketSortV}

end.